/**
 * Created on 18.05.2006
 * @author PG3D modded by Lars Lttmann
 */
package big3d.camera;

import big3d.utils.GLVector;
import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLU;

public class Camera {
    /**
     * Gibt die die Kamera-Transformation an
     */
    public Transformation camTransform;

    /**
     * ModelView-Matrix als <code>double[16]</code> zur Berechnung von Maus-
     * zu Weltkoordinaten,die einzelnen Zeilen der ModelView-Matrix von OpenGL
     * werden hintereinander in dieses Array geschrieben.
     */
    double[] modelViewMatrix;

    /**
     * Projektions-Matrix als <code>double[16]</code> zur Berechnung von Maus-
     * zu Weltkoordinaten, die einzelnen Zeilen der ModelView-Matrix von OpenGL
     * werden hintereinander in dieses Array geschrieben.
     */
    double[] projectionMatrix;

    /**
     * Array mit den Indizces der einzelnen Ansichten
     */
    int[] viewport;

    /**
     * X-Koordinate des Objekts
     */
    double[] objx;

    /**
     * Y-Koordinate des Objekts
     */
    double[] objy;

    /**
     * Z-Koordinate des Objekts
     */
    double[] objz;

    /**
     * Weltkoordinaten
     */
    double[] wcord;

    // GL-Sachen:
    GLDrawable gld;

    GL gl;

    GLU glu;

    // Position of the Camera in the fixed axis system
    protected GLVector position = new GLVector();

    /**
     * Orientation of the camera Store x, y, z (ignored) rotation angles. y
     * angle : turn around upward vector (ie look toward your sides). x angle :
     * turn around sidewward vector (ie look toward the ground or the sky). z
     * angle : ignored
     */
    protected GLVector orientation = new GLVector();

    public static Camera instance = null;


    private Camera(GLDrawable gld) {
        camTransform = new Transformation();

        this.modelViewMatrix = new double[16];
        this.projectionMatrix = new double[16];

        viewport = new int[4];

        objx = new double[1];
        objy = new double[1];
        objz = new double[1];


        this.gld = gld;
        this.gl = gld.getGL();
        this.glu = gld.getGLU();
    }

    /**
     * Liefert die aktuellen Weltkoordinaten dieser Kamera zur&uuml;ck.
     * 
     * <p>
     * Wenn die Kamera mit der Maus im Editor bewegt wird, so werden die
     * Maus-Koordinaten in die Berechnung der Kamera-Position in der Welt
     * hinzugezogen.
     * 
     * @param x
     *            X-Koordinate der Maus
     * @param y
     *            Y-Koordinate der Maus
     * @return Das <code>double[]</code> mit den Welt-Koordinaten
     */
    public double[] getWorldCoords(int x, int y) {
        double winX = (double) x;
        double winY = (double) viewport[3] - (double) y;

        float[] winZ = new float[1];
        gl.glReadPixels((int) winX, (int) winY, 1, 1, GL.GL_DEPTH_COMPONENT,
                GL.GL_FLOAT, winZ);

        glu.gluUnProject(winX, winY, (double) winZ[0], this.modelViewMatrix,
                this.projectionMatrix, viewport, objx, objy, objz);

        double[] result = { objx[0], objy[0], objz[0] };
        return result;
    }

    /**
     * Wendet die Fenster-Transformation auf die Kamera an.
     */
    public void applyFrameTransform() {
        gl.glMatrixMode(GL.GL_MODELVIEW);

        this.camTransform.composeGLMatrix();
        gl.glLoadMatrixf(this.camTransform.m_matrix);

        // fr glu-unproject mssen diese zeilen exakt hier stehen.
        gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, this.modelViewMatrix);
        gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, this.projectionMatrix);
        gl.glGetIntegerv(GL.GL_VIEWPORT, this.viewport);

        
    }

    /**
     * Verschiebt die Kamera.
     * 
     * <p>
     * Die Transformation der Kamera ist invers zu Objekt-Transformationen.
     * </p>
     * 
     * @param x
     *            Wert der Verschiebung in globale x-Richtung
     * @param y
     *            Wert der Verschiebung in globale y-Richtung
     * @param z
     *            Wert der Verschiebung in globale z-Richtung
     */
    public void translateGlobal(float x, float y, float z) {

        Transformation t = new Transformation();
        t.translate3f(-x, -y, -z);

        // reihenfolge hier: A, B -> erst B, darauf A
        this.camTransform.computeProduct(this.camTransform, t);
    }

    /**
     * Verschiebt die Kamera in lokalem Koordinatensystem.
     * <p>
     * Die Transformation der Kamera ist invers zu Objekt-Transformationen.
     * </p>
     * 
     * @param x
     *            Wert der Verschiebung in lokale x-Richtung
     * @param y
     *            Wert der Verschiebung in lokale y-Richtung
     * @param z
     *            Wert der Verschiebung in lokale z-Richtung
     */
    public void translateLocal(float x, float y, float z) {
        Transformation t = new Transformation();
        t.translate3f(-x, -y, -z);

        // reihenfolge hier: A, B -> erst B, darauf A
        this.camTransform.computeProduct(t, this.camTransform);
    }

    /**
     * Legt die Translationwerte fest.
     * 
     * @param x
     *            x-Wert der Translation
     * @param y
     *            y-Wert der Translation
     * @param z
     *            z-Wert der Translation
     */
    public void setTranslate(float x, float y, float z) {
        this.camTransform.setTranslate3f(-x, -y, -z);
    }

    /**
     * Dreht die Kamera um die Achsen des globalen Koordinatensystems.
     * 
     * <p>
     * Die Transformation der Kamera ist invers zu Objekt-Transformationen.
     * </p>
     * 
     * @param x
     *            Drehung um X-Achse
     * @param y
     *            Drehung um Y-Achse
     * @param z
     *            Drehung um Z-Achse
     */
    public void rotateGlobal(float x, float y, float z) {

        Transformation t = new Transformation();
        t.rotate3f((float) Math.toRadians(-x), (float) Math.toRadians(-y),
                (float) Math.toRadians(-z));

        // reihenfolge hier: A, B -> erst B, darauf A
        this.camTransform.computeProduct(this.camTransform, t);

    }

    /**
     * Dreht die Kamera um die Achsen des lokalen Koordinatensystems.
     * 
     * <p>
     * Die Transformation der Kamera ist invers zu Objekt-Transformationen.
     * </p>
     * 
     * @param x
     *            Drehung um X-Achse
     * @param y
     *            Drehung um Y-Achse
     * @param z
     *            Drehung um Z-Achse
     */
    public void rotateLocal(float x, float y, float z) {

        Transformation t = new Transformation();
        t.rotate3f((float) Math.toRadians(-x), (float) Math.toRadians(-y),
                (float) Math.toRadians(-z));
        // reihenfolge hier: A, B -> erst B, darauf A
        this.camTransform.computeProduct(t, this.camTransform);
    }

    /**
     * Setzt die Modelview-Matrix auf die &uuml;bergebene Projektion und wirft
     * die vorige Transformation auf den Matrix-Stack.
     * 
     * @param t
     *            zu setzende <code>Transformation</code>
     */
    public void setTransformation(Transformation t) {
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();
        gl.glMultMatrixf(t.m_matrix);
    }

    /**
     * Stellt oberste <code>Transformation</code> auf dem Stack wieder her.
     */
    public void restoreTransformation() {
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPopMatrix();
    }
    
    /**
     * Liefert die aktuelle Position dieser Kamera als Array zur&uuml;ck.
     * 
     * <p>Die Koordinaten stehen in folgender Reihenfolge in dem <code>float</code>-Array:<br>
     * <code>[ X-Wert | Y-Wert | Z-Wert ]</code></p>
     * 
     * @return Position der Kamera als float[3].
     */
    public float[] getCamPosition() {
        //float[] pos = { this.m_positionX, this.m_positionY, this.m_positionZ };
        //return pos;
        return this.camTransform.getTranslate3fv();
    }
    
    /**
     * Legt lokale Rotationswinkel fest.
     * 
     * @param x lokaler x-Winkel
     * @param y lokaler y-Winkel
     * @param z lokaler z-Winkel
     */
    public void setRotateLocal(float x, float y, float z) {
        this.camTransform.setRotate3fRad(
                (float) Math.toRadians(-x),
                (float) Math.toRadians(-y),
                (float) Math.toRadians(-z));                
    }
   

    public static Camera getInstance(GLDrawable gld) {
        if (instance == null)
            instance = new Camera(gld);
        return instance;
    }

    public static Camera getInstance() {
        if (instance == null)
            return null;
        return instance;
    }


}
