/**
 * Created on 18.05.2006
 * @author PG3D modded by Lars Lttmann
 */
package big3d.camera;

import big3d.utils.MathHelper;

public class Transformation {
    /**
     * colum-major Matrix
     */
    public float m_matrix[] = new float[16]; // column-major.   
    /**
     * 3x3-Rotationsmatrix als 9-stelliger Vektor
     */
    public float m_rotate[] = new float[9]; // column-major.
    /**
     * Radianter X-Winkel
     */
    public float rotX;
    /**
     * Radianter Y-Winkel
     */
    public float rotY;
    /**
     * Radianter Z-Winkel
     */
    public float rotZ; // in Rad !
    /**
     * der 3-stellige Translationsvektor
     */
    public float m_translate[] = new float[3];
    /**
     * Flag, ob diese Transformation die Identit&auml;t ist oder nicht
     */
    private boolean m_isIdentity;   
   
    /**
     * Konstruktor.
     * 
     * <p>Initialisierung einer uniformen Transformation, 
     * die auch die Identit&auml;t ist.</p>
     */
    public Transformation()
    {
        m_isIdentity = true;
        setIdentity();            

        composeGLMatrix();
    }
    

    /**
     * Copy-Konstruktor.
     * 
     * @param transf die zu kopierende Transformation
     */
    public Transformation(Transformation transf)
    {   
        m_isIdentity = transf.isIdentity();
         
        rotX = transf.getRotationRadX();
        rotY = transf.getRotationRadY();
        rotZ = transf.getRotationRadZ();
        
        for (int i = 0; i < 9; i++){
            m_rotate[i] = transf.m_rotate[i];
        }
        m_translate[0] = transf.m_translate[0];
        m_translate[1] = transf.m_translate[1];
        m_translate[2] = transf.m_translate[2];
 
       this.composeGLMatrix();
        
    }
    
    
    /**
     * Legt die Identit&auml;t dieser Transformation fest.
     */
    public void setIdentity()
    {
        rotX=0.0f;
        rotY=0.0f;
        rotZ=0.0f;
        
        m_translate[0]=0.0f;
        m_translate[1]=0.0f;
        m_translate[2]=0.0f;
        
        // x-spalte
        m_rotate[0] = 1.0f;
        m_rotate[1] = 0.0f;
        m_rotate[2] = 0.0f;
        
        m_matrix[0] = 1.0f;
        m_matrix[1] = 0.0f;
        m_matrix[2] = 0.0f;
        m_matrix[3] = 0.0f;
        
        // y-spalte
        m_rotate[3] = 0.0f;
        m_rotate[4] = 1.0f;
        m_rotate[5] = 0.0f;
        
        m_matrix[4] = 0.0f;
        m_matrix[5] = 1.0f;
        m_matrix[6] = 0.0f;
        m_matrix[7] = 0.0f;
        
        // z-spalte
        m_rotate[6] = 0.0f;
        m_rotate[7] = 0.0f;
        m_rotate[8] = 1.0f;               

        m_matrix[8] = 0.0f;
        m_matrix[9] = 0.0f;
        m_matrix[10] = 1.0f;
        m_matrix[11] = 0.0f;
        
        m_matrix[12] = 0.0f;
        m_matrix[13] = 0.0f;
        m_matrix[14] = 0.0f;
        m_matrix[15] = 1.0f;

        composeGLMatrix();
        m_isIdentity = true;
    }

    
    /**
     * Pr&uuml;ft, ob die Transformation die Identit&auml;t ist.
     * 
     * @return <code>true</code>, wenn es die Identit&auml;t ist, sonst 
     *         <code>false</code>
     */
    public boolean isIdentity()
    {
        return m_isIdentity;
    }
    
    /**
     * Setzt die Rotationsmatrix dieser Transformation zur&uuml;ck.
     */
    public void clearRotate()
    {
        rotX = 0.0f;
        rotY = 0.0f;
        rotZ = 0.0f;
        
        // x-spalte
        m_rotate[0] = 1.0f;
        m_rotate[1] = 0.0f;
        m_rotate[2] = 0.0f;
        
        // y-spalte
        m_rotate[3] = 0.0f;
        m_rotate[4] = 1.0f;
        m_rotate[5] = 0.0f;
        
        // z-spalte
        m_rotate[6] = 0.0f;
        m_rotate[7] = 0.0f;
        m_rotate[8] = 1.0f;
        
        composeGLMatrix();
    }
    /**
     * Generiert eine <code>OpenGL</code> verst&auml;dliche 4x4 Matrix ohne
     * Transformationen. (fr die SkyBox wichtig!)
     */
    public void composeGLMatrixNoTransformation(){
//      nun die 4x4matrix bilden:
        m_matrix[0] = m_rotate[0];
        m_matrix[1] = m_rotate[1];
        m_matrix[2] = m_rotate[2];       
        m_matrix[3] = 0.0f;
                
        m_matrix[4] = m_rotate[3];
        m_matrix[5] = m_rotate[4];
        m_matrix[6] = m_rotate[5];       
        m_matrix[7] = 0.0f;
    
        m_matrix[8] = m_rotate[6];
        m_matrix[9] = m_rotate[7];
        m_matrix[10]= m_rotate[8];       
        m_matrix[11]= 0.0f;
    
        m_matrix[12]= 0.0f;
        m_matrix[13]= 0.0f;
        m_matrix[14]= 0.0f;       
        m_matrix[15]= 1.0f;
    }
    
    /**
     * Legt neue Rotationswinkel der Transformation fest, erwartet drei 
     * Winkel-Angaben (Radiant).
     * 
     * @param xAngle Winkel der X-Achse
     * @param yAngle Winkel der Y-Achse
     * @param zAngle inkel der Z-Achse
     */
    public void setRotate3fRad (float xAngle, float yAngle, float zAngle)
    {             
        clearRotate();
        rotate3f(xAngle, yAngle, zAngle, false);
        composeGLMatrix();
    }
  
  
    /**
     * Rotiert die Transformation, erwartet drei Winkel-Angaben (Grad).
     * 
     * @param xAngle Winkel der X-Achse [0.0, 360.0]
     * @param yAngle Winkel der Y-Achse [0.0, 360.0]
     * @param zAngle Winkel der Z-Achse [0.0, 360.0]
     */
    public void setRotate3fDeg (float xAngle, float yAngle, float zAngle)
    {             
        clearRotate();
        rotate3f(
                (float) Math.toRadians(xAngle),
                (float) Math.toRadians(yAngle),
                (float) Math.toRadians(zAngle));
        composeGLMatrix();
    }

    /**
     * Rotiert die Transformation, erwartet drei Winkel-Angaben (Grad).
     * 
     * @param xAngle Winkel der X-Achse [0.0, 360.0]
     * @param yAngle Winkel der Y-Achse [0.0, 360.0]
     * @param zAngle Winkel der Z-Achse [0.0, 360.0]
     * @param invertorder Flag, ob invertiert werden soll
     */
    public void setRotate3fDeg (float xAngle, float yAngle, float zAngle, boolean invertorder)
    {             
        clearRotate();
        rotate3f(
                (float) Math.toRadians(xAngle),
                (float) Math.toRadians(yAngle),
                (float) Math.toRadians(zAngle), invertorder);
        composeGLMatrix();
    }


   
    /**
     * Rotiert die Transformation, erwartet drei Winkel-Angaben.
     * 
     * @param xAngle Winkel der X-Achse
     * @param yAngle Winkel der Y-Achse
     * @param zAngle Winkel der Z-Achse
     */
    public void rotate3f (float xAngle, float yAngle, float zAngle)
    {   
        rotate3f (xAngle, yAngle, zAngle, false);   
    }

    /**
     * Rotiert die Transformation, erwartet drei Winkel-Angaben (Grad).
     * 
     * @param xAngle Winkel der X-Achse [0.0, 360.0]
     * @param yAngle Winkel der Y-Achse [0.0, 360.0]
     * @param zAngle Winkel der Z-Achse [0.0, 360.0]
     */
    public void rotate3fDeg (float xAngle, float yAngle, float zAngle)
    {   
        rotate3f ((float) Math.toRadians(xAngle), (float) Math.toRadians(yAngle), (float) Math.toRadians(zAngle), false);   
    }
    /**
     * Rotiert die Transformation, erwartet drei Winkel-Angaben (Grad).
     * 
     * @param xAngle Winkel der X-Achse [0.0, 360.0]
     * @param yAngle Winkel der Y-Achse [0.0, 360.0]
     * @param zAngle Winkel der Z-Achse [0.0, 360.0]
     * @param invertOrder Flag, ob invertiert werden soll
     */
    public void rotate3fDeg (float xAngle, float yAngle, float zAngle, boolean invertOrder)
    {   
        rotate3f ((float) Math.toRadians(xAngle), (float) Math.toRadians(yAngle), (float) Math.toRadians(zAngle), invertOrder);   
    }    
    
    /**
     * Rotiert die Transformation, erwartet drei Winkel-Angaben (Radiant) 
     * sowie ein Invertierungs-Flag.
     * 
     * @param xAngle      Winkel der X-Achse
     * @param yAngle      Winkel der X-Achse
     * @param zAngle      Winkel der X-Achse
     * @param invertOrder <code>true</code>, wenn invertiert werden soll,
     *                    sonst <code>false</code>
     */
    public void rotate3f (float xAngle, float yAngle, float zAngle, boolean invertOrder)
    {       
   
        float cos, sin;   
        this.rotX += xAngle; 
        this.rotY += yAngle;
        this.rotZ += zAngle;
        
        // zuerst die drei rotationsmatritzen fr die achsen bauen:
        cos = (float)Math.cos(xAngle);
        sin = (float)Math.sin(xAngle);
        float xMatrix[] = {1.0f, 0.0f, 0.0f,
                   0.0f, cos, -sin, 
                   0.0f, sin, cos};            
        
        cos = (float)Math.cos(yAngle);
        sin = (float)Math.sin(yAngle);              
        float yMatrix[] =  {
                cos, 0.0f, sin,
                0.0f, 1.0f, 0.0f,
                -sin, 0.0f, cos };
        
        cos = (float)Math.cos(zAngle);
        sin = (float)Math.sin(zAngle);
        float zMatrix[] = {
            cos, -sin, 0.0f,
            sin, cos, 0.0f, 
            0.0f , 0.0f, 1.0f };        
             
        //nun rechnen wir die drei einzelnen matritzen auf die matrix drauf:
        
        if(!invertOrder)
        {
            float temp[]  = MultMatrixMatrix3(xMatrix, m_rotate);
            float temp2[] = MultMatrixMatrix3(zMatrix, yMatrix);          
            this.m_rotate = MultMatrixMatrix3( temp2, temp);
        }
        else
        {
            float temp[]  = MultMatrixMatrix3(yMatrix, zMatrix);          
            float temp2[] = MultMatrixMatrix3(m_rotate, xMatrix);
            this.m_rotate = MultMatrixMatrix3( temp2, temp);            
        }
    
        m_isIdentity = false;
        composeGLMatrix();
    }

    
    /**
     * Liefert den Rotationswinkel um die X-Achse (RADIANT) dieser
     * Transformation zur&uuml;ck.
     * 
     * @return den Rotationswinkel (RADIANT)
     */
    public float getRotationRadX()
    {
        return rotX;
    }

   
    /**
     * Liefert den Rotationswinkel um die X-Achse (in GRAD) dieser
     * Transformation zur&uuml;ck.
     * 
     * @return den Rotationswinkel (Gradangabe, zwischen 0 und 360)
     */
    public float getRotationDegX()
    {
        return (float) Math.toDegrees(rotX);
    }

    /**
     * Liefert den Rotationswinkel um die Y-Achse (RADIANT) dieser
     * Transformation zur&uuml;ck.
     * 
     * @return den Rotationswinkel (RADIANT)
     */
    public float getRotationRadY()
    {
        return rotY;
    }
    
  
    /**
     * Liefert den Rotationswinkel um die Y-Achse (in GRAD) dieser
     * Transformation zur&uuml;ck.
     * 
     * @return den Rotationswinkel (Gradangabe, zwischen 0 und 360)
     */
    public float getRotationDegY()
    {
        return (float) Math.toDegrees(rotY);
    }

    
    /**
     * Liefert den Rotationswinkel um die Z-Achse (RADIANT) dieser
     * Transformation zur&uuml;ck.
     * 
     * @return den Rotationswinkel (RADIANT)
     */
    public float getRotationRadZ()
    {
        return rotZ;
    }
    
    
    /**
     * Liefert den Rotationswinkel um die Z-Achse (in GRAD) dieser
     * Transformation zur&uuml;ck.
     * 
     * @return den Rotationswinkel (Gradangabe, zwischen 0 und 360)
     */
    public float getRotationDegZ()
    {
        return (float) Math.toDegrees(rotZ);
    }


    /**
     * Liefert den aktuellen Rotationsvektor der Rotation um die X-Achse
     * (vom Typ <code>float[3]</code>) dieser Transformation zur&uuml;ck.
     * 
     * @return den Rotationsvektor
     */
    public float[] getXRotationVec3fv()
    {
        float[] ret = new float[3];        
        ret[0] = m_rotate[0];
        ret[1] = m_rotate[1];
        ret[2] = m_rotate[2];
        return ret;
    }
    
 
    /**
     * Liefert den aktuellen Rotationsvektor der Rotation um die Y-Achse
     * (vom Typ <code>float[3]</code>) dieser Transformation zur&uuml;ck.
     * 
     * @return den Rotationsvektor
     */
    public float[] getYRotationVec3fv()
    {
        float[] ret = new float[3];        
        ret[0] = m_rotate[3];
        ret[1] = m_rotate[4];
        ret[2] = m_rotate[5];
        return ret;
    }

   
    /**
     * Liefert den aktuellen Rotationsvektor der Rotation um die Z-Achse
     * (vom Typ <code>float[3]</code>) dieser Transformation zur&uuml;ck.
     * 
     * @return den Rotationsvektor
     */
    public float[] getZRotationVec3fv()
    {
        float[] ret = new float[3];        
        ret[0] = m_rotate[6];
        ret[1] = m_rotate[7];
        ret[2] = m_rotate[8];
        return ret;
    }
    
    
    /**
     * Legt einen neuen 3-stelligen Translationvektor f&uuml;r diese
     * Transformation fest.
     * 
     * @param translate 3-stelliger Translationvektor
     */
    public void setTranslate3fv (float translate[])
    {
        m_translate[0]=translate[0];
        m_translate[1]=translate[1];
        m_translate[2]=translate[2];
        m_isIdentity = false;

        m_matrix[12]= m_translate[0];
        m_matrix[13]= m_translate[1];
        m_matrix[14]= m_translate[2];   
    }

    /**
     * Legt einen neuen 3-stelligen Translationvektor, repr&auml;sentiert
     * durch die &uuml;bergebenen Parameter, f&uuml;r diese
     * Transformation fest.
     * 
     * @param x  der x-Wert der Translation
     * @param y  der y-Wert der Translation
     * @param z  der z-Wert der Translation
     */
    public void setTranslate3f (float x, float y, float z)
    {        
        m_translate[0]=x;
        m_translate[1]=y;
        m_translate[2]=z;
                
        m_isIdentity = false;

        m_matrix[12]= m_translate[0];
        m_matrix[13]= m_translate[1];
        m_matrix[14]= m_translate[2];   
    }

    /**
     * Liefert den aktuellen 3-stelligen Translationsvektor dieser 
     * Transformation zur&uuml;ck.
     * 
     * @return der Translationsvektor
     */
    public float[] getTranslate3fv ()
    {
        float[] ret = new float[3];
        ret[0] = m_translate[0];
        ret[1] = m_translate[1];
        ret[2] = m_translate[2];

        return ret;
    }

    /**
     * Liefert den aktuellen 4-stelligen Translationsvektor dieser 
     * Transformation zur&uuml;ck.
     * 
     * @return der Translationsvektor
     */
    public float[] getTranslate4fv ()
    {
        float[] ret = new float[4];
        ret[0] = m_translate[0];
        ret[1] = m_translate[1];
        ret[2] = m_translate[2];
        ret[3] = 1f;

        return ret;
    }

    /**
     * Transloziert diese Transformation.
     * 
     * <p>Diese Methode erwartet 3 Parameter, die als <code>float</code>
     * -Werte &uuml;bergeben werden.</p>
     * 
     * @param x  die Translation entlang der x-Achse
     * @param y  die Translation entlang der y-Achse
     * @param z  die Translation entlang der z-Achse
     */
    public void translate3f (float x, float y, float z)
    {       
        m_translate[0]+=x;
        m_translate[1]+=y;
        m_translate[2]+=z;
        
        m_isIdentity = false;

        m_matrix[12]= m_translate[0];
        m_matrix[13]= m_translate[1];
        m_matrix[14]= m_translate[2];   
    }
    
    // Bewegt das Objekt entlang seiner lokalen Achsen
    /**
     * Bewegt das Objekt entlang der lokalen Achsen.
     * 
     * @param x  die Translation entlang der x-Achse
     * @param y  die Translation entlang der y-Achse
     * @param z  die Translation entlang der z-Achse
     */
    public void relativTranslate3f (float x, float y, float z) {   
        Transformation t = new Transformation();
        t.setTranslate3f(x, y, z);
        this.m_matrix = MathHelper.mTransMultTrans(this.m_matrix, t.m_matrix);
     
        this.reconstructTransformations();
    }

    /**
     * Transloziert diese Transformation.
     * 
     * <p>Diese Methode erwartet ein Array als Parameter, welches vom Typ
     * <code>float[3]</code> sein muss.</p>
     * 
     * @param translate 3-stelliger Translationsvektor
     */
    public void translate3fv (float translate[])
    {
        m_translate[0]+=translate[0];
        m_translate[1]+=translate[1];
        m_translate[2]+=translate[2];
        
        m_isIdentity = false;
        
        m_matrix[12]= m_translate[0];
        m_matrix[13]= m_translate[1];
        m_matrix[14]= m_translate[2];   
    }

    
    /**
     * Multipliziert zwei Rotationsmatrizen miteinander.
     * 
     * @param transA  die erste Transformation
     * @param transB  die zweite Transformation
     * @return das Ergebnis der Multiplikation als Vektor
     */
    public static float[] MultMatrixMatrix3(float[] transA, float[] transB)
    {
        float[] temp = new float[9];
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                for (int k = 0; k <3; k++) {
                    temp[i + (j*3)] += transA[i+(k*3)] * transB[k+(j*3)];
                }
            }
        }
        return temp;      
    }    

    /**
     * Multipliziert eine Matrix mit einem 3-stelligen Vektor.
     * 
     * @param transA die Matrix
     * @param b      der Vektor
     * @return das Ergebnis der Multiplikation
     */   
    public static float[] MultMatrixVector3(float[] transA, float[] b) {
        float[] temp = new float[3];
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                temp[i] += transA[i+(3*j)] * b[j];
            }
        }
        return temp;
    }
    
    
    /**
     * Multipliziert zwei Transformationen miteinander.
     * 
     * @param rkA die erste Transformation
     * @param rkB die zweite Transformation
     */
    
    public void computeProduct(Transformation rkA, Transformation rkB)
    {
       
        // neu: die rotation wird nur als 3x3 verwaltet. hier wird
        // sie zusammengerechnet:
        this.m_rotate = MultMatrixMatrix3(rkA.m_rotate, rkB.m_rotate);
        
        // nun die transformation. dorthinein mssen wir die transformation des
        // elternteils hineinrehcnen, denn sonst rotieren sich alle kinder eines
        // knoten nur um sich selbst, aber nicht um die position der gruppe:  
        float[] temp =  MultMatrixVector3(rkA.m_rotate, rkB.m_translate);
        setTranslate3fv(temp);
        translate3fv(rkA.m_translate);
                
        // nun noch eine OpenGL-konforme Matrix herstellen:
        composeGLMatrix();
        
    }

    /**
     * Bildet die 4x4-GLMatrix aus der kleinen 3x3-Rotationsmatrix und dem
     * Translationsvektor.
     */
    public void composeGLMatrix()
    {        
        // nun die 4x4matrix bilden:
        m_matrix[0] = m_rotate[0];
        m_matrix[1] = m_rotate[1];
        m_matrix[2] = m_rotate[2];       
        m_matrix[3] = 0.0f;
                
        m_matrix[4] = m_rotate[3];
        m_matrix[5] = m_rotate[4];
        m_matrix[6] = m_rotate[5];       
        m_matrix[7] = 0.0f;
    
        m_matrix[8] = m_rotate[6];
        m_matrix[9] = m_rotate[7];
        m_matrix[10]= m_rotate[8];       
        m_matrix[11]= 0.0f;
    
        m_matrix[12]= m_translate[0];
        m_matrix[13]= m_translate[1];
        m_matrix[14]= m_translate[2];       
        m_matrix[15]= 1.0f;
    }
    
    // von malte: rekonstruiert die kleine 3x3-rotate-matrix und den translate-vektor aus der 4x4GLMatrix
    /**
     * Rekonstruiert die kleine 3x3-Rotationsmatrix und den 
     * Translationsvektor aus der 4x4-GLMatrix.
     */
    public void reconstructTransformations(){
        
        m_rotate[0] = m_matrix[0];
        m_rotate[1] = m_matrix[1];
        m_rotate[2] = m_matrix[2];   
                
        m_rotate[3] = m_matrix[4];
        m_rotate[4] = m_matrix[5];
        m_rotate[5] = m_matrix[6]; 
    
        m_rotate[6] = m_matrix[8];
        m_rotate[7] = m_matrix[9];
        m_rotate[8] = m_matrix[10];  
    
        m_translate[0] = m_matrix[12];
        m_translate[1] = m_matrix[13];
        m_translate[2] = m_matrix[14];  
        
    }
    
    
  
    /**
     * Liefert eine mittels Gaussverfahren erstellte Inverse Matrix/Transformation.
     * 
     * <p>Leider stehen bei globalen Transformationsobjekten keine Rotationsangaben 
     * pro Achse zur Verfuegung, so dass jetzt die LGS-Klasse bemueht wird.
     * Im Fehlerfall wird die Einheitsmatrix erstellt.
     * Das Ergebnis selbst hat auch keine Rotationsangben pro Achse.</p>
     * 
     * @return Inverse Transformation
     */
    public Transformation getInvers() {
        Transformation t = new Transformation(this);
        
        try {
            float[][] workingMatrix = new float[][] {
                    { m_matrix[ 0], m_matrix[ 4], m_matrix[ 8], m_matrix[12] },
                    { m_matrix[ 1], m_matrix[ 5], m_matrix[ 9], m_matrix[13] },
                    { m_matrix[ 2], m_matrix[ 6], m_matrix[10], m_matrix[14] },
                    { m_matrix[ 3], m_matrix[ 7], m_matrix[11], m_matrix[15] } };

            float[][] column1 = MathHelper.gauss(workingMatrix, 
                    new float[] { 1, 0, 0, 0 });
            float[][] column2 = MathHelper.gauss(workingMatrix, 
                    new float[] { 0, 1, 0, 0 });
            float[][] column3 = MathHelper.gauss(workingMatrix, 
                    new float[] { 0, 0, 1, 0 });
            float[][] column4 = MathHelper.gauss(workingMatrix, 
                    new float[] { 0, 0, 0, 1 });

            t.m_matrix[0] = column1[0][0];
            t.m_matrix[1] = column1[1][0];
            t.m_matrix[2] = column1[2][0];
            t.m_matrix[3] = column1[3][0];

            t.m_matrix[4] = column2[0][0];
            t.m_matrix[5] = column2[1][0];
            t.m_matrix[6] = column2[2][0];
            t.m_matrix[7] = column2[3][0];

            t.m_matrix[ 8] = column3[0][0];
            t.m_matrix[ 9] = column3[1][0];
            t.m_matrix[10] = column3[2][0];
            t.m_matrix[11] = column3[3][0];

            t.m_matrix[12] = column4[0][0];
            t.m_matrix[13] = column4[1][0];
            t.m_matrix[14] = column4[2][0];
            t.m_matrix[15] = column4[3][0];
        
            t.reconstructTransformations();

        } catch (Exception e) {
            t.setIdentity();
        }
        return t;
    }
    
    /**
     * Wendet die Transformationsmatrix auf einen Vektor (i.d.R. Vertex.data) an.
     * 
     * @param v float[3]
     */
    public void transformVectorDirect(float[] v) {
        float[] temp = transformVector(v);
        v[0] = temp[0];
        v[1] = temp[1];
        v[2] = temp[2];
    }
    
    /**
     * Wendet die Transformationsmatrix auf einen Vektor (i.d.R. Vertex.data) an.
     * 
     * @param v float[3]
     * @return Transformierter Vektor (float[3])
     */
    public float[] transformVector(float[] v) {
        float[] temp = new float[]{0f, 0f, 0f};
        for(int i = 0; i < 3; i++) {
            for(int j = 0; j < 3; j++) {
                temp[i] += v[j] * this.m_matrix[i + (4*j)];
            }
            temp[i] += this.m_matrix[i + 12];
        }
        return temp;
    }

}
