package big3d.objects;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLU;
import net.java.games.jogl.util.GLUT;


/**
 * <b>LightObject</b><br><br>
 * <code>public class <b>LightObject</b></code><br><br>
 * Diese Klasse realisiert das eigentliche Licht-Objekt mit s&auml;mtlichen Licht-
 * Eigenschaften und die Festlegung der Licht-Art (ambient, diffus, spekular,
 * emissional).<br>
 * Es k&ouml;nnen maximal <b>acht</b> Lichtobjekte in einer Welt existieren.<br>
 * <i>Ambient</i> bedeutet...<br>
 * <i>Diffus</i> bedeutet...<br>
 * <i>Spekular</i> bedeutet...<br>
 * <i>Emissional</i> bedeutet...<br>
 * <br><b>Globale Variablen</b>:<br>
 * <code>private float[] <b>ambient</b></code> - Werte f&uuml;r ambientes Licht (RGBA)<br>
 * <code>private float[] <b>diffuse</b></code> - Werte f&uuml;r diffuses Licht (RGBA)<br>
 * <code>private float[] <b>specular</b></code> - Werte f&uuml;r speculares Licht (RGBA)<br>
 * <code>private float[] <b>position</b></code> - Werte f&uuml;r die Position (XYZ w)<br>
 * <code>private float[] <b>spotDirection</b></code> - Werte f&uuml;r die Spot-Richtung<br>
 * <code>private float <b>spotExponent</b></code> - Wert f&uuml;r die Spothelligkeit<br>
 * <code>private float <b>spotCutoff</b></code> - Wert f&uuml;r den Winkel<br>
 * <code>private float <b>constantAttenuation</b></code> - konstante Attenuation<br>
 * <code>private float <b>linearAttenuation</b></code> - lineare Attenuation<br>
 * <code>private float <b>quadraticAttenuation</b></code> - quadratische Attenuation<br>
 * <code>private float <b>spot</b></code> - Hilfsvariable zur Speicherung der Position des Spots<br>
 *  gemacht wird</code>
 * @author Florian Hinz, Lars L&uuml;ttmann, Malte Mathiszig
 */
public class LightObject{
    private static int openGLLightID = GL.GL_LIGHT0;
    private int myGLLightID = -1;
    
	private float[] ambient = { 0.4f, 0.4f, 0.4f, 1.0f };
	private float[] diffuse = { 0.5f, 0.5f, 0.5f, 1.0f };
	private float[] specular = { 0.2f, 0.2f, 0.2f, 1.0f };
	private float[] position = { 0.0f, 0.0f, 0.0f, 1.0f };
	private float[] spotDirection = { 0.0f, 0.0f, -1.0f };
	private float spotExponent = 15.0f;
	private float spotCutoff = 180.0f;
	private float constantAttenuation = 1.0f;
	private float linearAttenuation = 0.0f;
	private float quadraticAttenuation = 0.0f;
	
	protected static GLUT glut = new GLUT();;

    
    /* ---------- Initialisierung ---------- */
    /**
     * <b>LightObject</b> - Konstruktor<br><br>
     * <code>public <b>LightObject</b>()</code><br><br>
     * <code>default</code>-Initialisierung des Licht-Objekts. 
     */
	public LightObject() {
	}
    
    public LightObject (LightObject other){
        this.spotExponent = other.spotExponent;
        this.constantAttenuation = other.constantAttenuation;
        this.linearAttenuation = other.linearAttenuation;
        this.quadraticAttenuation = other.quadraticAttenuation;
        this.spotCutoff = other.spotCutoff;
        
        for (int i=0; i<4; i++){
            this.ambient[i] = other.ambient[i];
            this.diffuse[i] = other.diffuse[i];
            this.specular[i] = other.specular[i];
            this.position[i] = other.position[i];
        }
        
        for(int i=0; i<3; i++){
            this.spotDirection[i] = other.spotDirection[i];
        }
    }
	
	/* ---------- Zeichen-Methoden ---------- */

    
    /**
     * Diese Draw-Methode zeichnet eine grafische, kugel- oder kegelfrmige
     * Reprsentation des Lichtobjekts.
     * 
     */
    public void draw(GLDrawable gld){
        GL gl = gld.getGL();
        GLU glu = gld.getGLU();
        
		gl.glColor3f(1.0f, 1.0f, 0f);
		
        //alte modelview sichern:
        gl.glPushMatrix();
        
        float x = 0;
        //verschieben der "welt" 
        gl.glTranslatef(this.position[0], this.position[1], this.position[2]);
       
		// Kugel zeichnen
		glut.glutSolidSphere(glu,2.0f,10,10);
        
        x = (x + 0.9f)%360;
    	
        //Modelview wiederherstellen:
        gl.glPopMatrix();
        gl.glColor3f(0f, 0f, 0f);
    }
	
	/* ---------- apply-Methoden ---------- */
    
    /**
     * <b>applyLightProperties</b><br>
     * <br>
     * <code>public void <b>applyLightProperties</b>(int light, GL gl)</code><br>
     * <br>
     * Diese Methode legt die Eigenschaften eines Lichtes bei <code>OpenGL</code>
     * fest.
     * 
     * @param light -
     *            Nummer des Lichtes.
     * @param gl -
     *            Ein Objekt vom Typ <code>GL</code>.
     */
    public void applyLightProperties(int light, GL gl) {
       
        if (this == null)
            return;
        
        gl.glLightfv(light, GL.GL_AMBIENT, this.getAmbient());
        gl.glLightfv(light, GL.GL_DIFFUSE, this.getDiffuse());
        gl.glLightfv(light, GL.GL_SPECULAR, this.getSpecular());
        gl.glLightfv(light, GL.GL_POSITION, this.getPosition());
    
        // Ist die vierte Komponente (w) des Lichtes gleich 0, so handelt
        // es sich um ein direktionales Licht und die folgenden Attribute
        // haben keinen Einfluss, mssen also auch nicht gesetzt werden.
        
      if (this.getPosition()[3] != 0.0f) {
            gl.glLightf(light, GL.GL_CONSTANT_ATTENUATION, this
                    .getConstantAttenuation());
            gl.glLightf(light, GL.GL_LINEAR_ATTENUATION, this
                    .getLinearAttenuation());
            gl.glLightf(light, GL.GL_QUADRATIC_ATTENUATION, this
                    .getQuadraticAttenuation());
            gl.glLightf(light, GL.GL_SPOT_CUTOFF, this.getSpotCutoff());
            gl.glLightf(light, GL.GL_SPOT_EXPONENT, this
                    .getSpotExponent());
            gl.glLightfv(light, GL.GL_SPOT_DIRECTION, this
                    .getSpotDirection());
        }
     
    }
    
    /* ---------- set-Methoden ---------- */
    
	/**
	 * <b>setAmbient</b><br><br>
     * <code>public void <b>setAmbient</b>(float r, float g, float b, float a)</code><br><br>
     * Diese Methode legt die Farbwerte f&uuml;r ambientes Licht fest.
	 * @param r - Rot-Wert
	 * @param g - Gr&uuml;n-Wert
	 * @param b - Blau-Wert
	 * @param a - Alpha-Wert
	 */
	public void setAmbient(float r, float g, float b, float a) {
		ambient[0] = r;
		ambient[1] = g;
		ambient[2] = b;
		ambient[3] = a;
	}
    
    public void setAmbient(float[] color){
        if (color.length==4){
            ambient[0] = color[0];
            ambient[1] = color[1];
            ambient[2] = color[2];
            ambient[3] = color[3];
        }
    }
	
	/**
	 * <b>setDiffuse</b><br><br>
     * <code>public void <b>setDiffuse</b>(float r, float g, float b, float a)</code><br><br>
     * Diese Methode legt die Farbwerte f&uuml;r diffuses Licht fest.
	 * @param r - Rot-Wert
	 * @param g - Gr&uuml;n-Wert
	 * @param b - Blau-Wert
	 * @param a - Alpha-Wert
	 */
	public void setDiffuse(float r, float g, float b, float a) {
		diffuse[0] = r;
		diffuse[1] = g;
		diffuse[2] = b;
		diffuse[3] = a;
	}
	/**
	 * <b>setSpecular</b><br><br>
     * <code>public void <b>setSpecular</b>(float r, float g, float b, float a)</code><br><br>
     * Diese Methode legt die Farbwerte f&uuml;r spekulares Licht fest.
	 * @param r - Rot-Wert
	 * @param g - Gr&uuml;n-Wert
	 * @param b - Blau-Wert
	 * @param a - Alpha-Wert
	 */
	public void setSpecular(float r, float g, float b, float a) {
		specular[0] = r;
		specular[1] = g;
		specular[2] = b;
		specular[3] = a;
	}
	/**
	 * <b>setPosition</b><br><br>
     * <code>public void <b>setPosition</b>(float x, float y, float z, float w)</code><br><br>
     * Diese Methode legt die Position des Lichts fest.
	 * @param x - X-Koordinate
	 * @param y - Y-Koordinate
	 * @param z - Z-Koordinate
	 * @param w - w-Wert
	 */
	public void setPosition4f(float x, float y, float z, float w) {
		position[0] = x;
		position[1] = y;
		position[2] = z;
		position[3] = w;
	}
    
    /**
     * <b>setPosition</b><br><br>
     * <code>public void <b>setPosition</b>(float[] pos)</code><br><br>
     * Diese Methode legt die Position des Lichts fest.
     * @param pos - neue Position als <code>float[4]</code>
     */
    public void setPosition4fv(float[] pos){
        if (pos.length == position.length){
            position[0] = pos[0];
            position[1] = pos[1];
            position[2] = pos[2];
        }
        
    }

	/**
	 * <b>setSpotDirection</b><br><br>
     * <code>public void <b>setSpotDirection</b>(float x, float y, float z)</code><br><br>
     * Diese Methode legt die Richtung des Spots fest.
	 * @param x - x-Richtung
	 * @param y - y-Richtung
	 * @param z - z-Richtung
	 */
	public void setSpotDirection(float x, float y, float z) {
		spotDirection[0] = x;
		spotDirection[1] = y;
		spotDirection[2] = z;
	}
    
    public void setSpotDirection3f(float[] direction){
        if (direction.length == this.spotDirection.length){
            for (int i=0; i<direction.length; i++){
                this.spotDirection[i] = direction[i];
            }
        }
        
    }
	/**
	 * <b>setSpotExponent</b><br><br>
     * <code>public void <b>setSpotExponent</b>(float exp)</code><br><br>
     * Diese Methode legt die Helligkeit f&uuml;r den Spot fest.
	 * @param exp - die Helligkeit
	 */
	public void setSpotExponent(float exp) {
		spotExponent = exp;
	}
	/**
	 * <b>setSpotCutoff</b><br><br>
     * <code>public void <b>setSpotCutoff</b>(float cutoff)</code><br><br>
     * Diese Methode legt den Spot-Winkel fest.
	 * @param cutoff - der Winkel
	 */
	public void setSpotCutoff(float cutoff) {
		spotCutoff = cutoff;
	}
	/**
	 * <b>setConstantAttenuation</b><br><br>
     * <code>public void <b>setConstantAttenuation</b>(float att)</code><br><br>
     * Diese Methode legt die konstante Attenuation fest.
	 * @param att - die Attenuation
	 */
	public void setConstantAttenuation(float att) {
		constantAttenuation = att;
	}
	/**
	 * <b>setLinearAttenuation</b><br><br>
     * <code>public void <b>setLinearAttenuation</b>(float att)</code><br><br>
     * Diese Methode legt die lineare Attenuation fest.
	 * @param att - die Attenuation
	 */
	public void setLinearAttenuation(float att) {
		linearAttenuation = att;
	}
	/**
	 * <b>setQuadraticAttenuation</b><br><br>
     * <code>public void <b>setQuadraticAttenuation</b>(float att)</code><br><br>
     * Diese Methode legt die quadratische Attenuation fest.
	 * @param att - die Attenuation
	 */
	public void setQuadraticAttenuation(float att) {
		quadraticAttenuation = att;
	}
	/**
	 * <b>setAttenuation</b><br><br>
     * <code>public void <b>setAttenuation</b>(float constant, float linear, float quadratic)</code><br><br>
     * Diese Methode legt alle Attenuationswerte fest.
	 * @param constant - konstante Attenuation
	 * @param linear - lineare Attenuation
	 * @param quadratic - quadratische Attenuation
	 */
	public void setAttenuation(float constant, float linear, float quadratic) {
		constantAttenuation = constant;
		linearAttenuation = linear;
		quadraticAttenuation = quadratic;
	}

    /**
     * <b>setSpot</b><br><br>
	 * <code>public void <b>setSpot</b>(float spot)</code><br><br>
	 * Diese Methode legt die Position des Spots in der Hilfsvariable fest.
     * @param spot - die neue Spot-Position
     */
    public void setSpot(float spot) {
        this.position[3] = spot;
    }
    
	/* ---------- get-Methoden ---------- */
	
    /**
     * <b>getAmbient</b><br><br>
     * <code>public float[] <b>getAmbient()</b>()</code><br><br>
     * Liefert die Farbwerte des ambienten Lichts zur&uuml;ck.
     * @return Farbwerte
     */
	public float[] getAmbient() {
		return ambient;
	}
	/**
	 * <b>getDiffuse</b><br><br>
     * <code>public float[] <b>getDiffuse</b>()</code><br><br>
     * Liefert die Farbwerte des diffusen Lichts zur&uuml;ck.
	 * @return Farbwerte
	 */
	public float[] getDiffuse() {
		return diffuse;
	}
	/**
	 * <b>getSpecular</b><br><br>
     * <code>public float[] <b>getSpecular</b>()</code><br><br>
     * Liefert die Farbwerte des specularen Lichts zur&uuml;ck.
	 * @return Farbwerte
	 */
	public float[] getSpecular() {
		return specular;
	}
	/**
	 * <b>getPosition</b><br><br>
     * <code>public float[] <b>getPosition</b>()</code><br><br>
     * Liefert die Position des Lichts zur&uuml;ck.
	 * @return Position
	 */
	public float[] getPosition() {
		return position;
	}
	/**
	 * <b>getSpotDirection</b><br><br>
     * <code>public float[] <b>getSpotDirection</b>()</code><br><br>
     * Liefert die Richtung des Spots zur&uuml;ck.
	 * @return Richtung
	 */
	public float[] getSpotDirection() {
		return spotDirection;
	}
	/**
	 * <b>getSpotExponent</b><br><br>
     * <code>public float <b>getSpotExponent</b>()</code><br><br>
     * Liefert die Helligkeit des Spots zur&uuml;ck.
	 * @return Helligkeit
	 */
	public float getSpotExponent() {
		return spotExponent;
	}
	/**
	 * <b>getSpotCutoff</b><br><br>
     * <code>public float <b>getSpotCutoff</b>()</code><br><br>
     * Liefert den Winkel des Spots zur&uuml;ck.
	 * @return Winkel
	 */
	public float getSpotCutoff() {
		return spotCutoff;
	}
	/**
	 * <b>getConstantAttenuation</b><br><br>
     * <code>public float <b>getConstantAttenuation</b>()</code><br><br>
     * Liefert die Werte der konstanten Attenuation zur&uuml;ck.
	 * @return konstante Attenuation
	 */
	public float getConstantAttenuation() {
		return constantAttenuation;
	}
	/**
	 * <b>getLinearAttenuation</b><br><br>
     * <code>public float <b>getLinearAttenuation</b>()</code><br><br>
     * Liefert die Werte der linearen Attenuation zur&uuml;ck.
	 * @return lineare Attenuation
	 */
	public float getLinearAttenuation() {
		return linearAttenuation;
	}
	/**
	 * <b>getQuadraticAttenuation</b><br><br>
     * <code>public float <b>getQuadraticAttenuation</b>()</code><br><br>
     * Liefert die Werte der quadratischen Attenuation zur&uuml;ck.
	 * @return quadratische Attenuation
	 */
	public float getQuadraticAttenuation() {
		return quadraticAttenuation;
	}
    
    /**
     * Schaltet das Licht ein. Anm.: da wir davon ausgehen, dass man 
     * insg nicht mehr als 8 Lichter verwenden will (kann!) wird das Lschen von 
     * Lichtern nicht untersttzt.
     * @param gl
     */
	public void turnOnLight(GL gl){
        //merkt sich die eigene GL-Light ID:
        if (this.myGLLightID == -1)
            this.myGLLightID = openGLLightID;
        
        //setzt die Lichtwerte:
        this.applyLightProperties(this.myGLLightID, gl);
        
        gl.glGetError();
        //einschalten:
        gl.glEnable(this.myGLLightID);
        if (gl.glGetError() != GL.GL_NO_ERROR)
            System.err.println("Fehler beim Einschalten des Lichtes!");
        
        //erhht den Lichtindex (von OpenGL):
        openGLLightID += 1;
        
    }
    
    /**
     * Schaltet das Licht aus.
     * @param gl
     */
    public void turnOffLight(GL gl){
         gl.glDisable(myGLLightID);
    }
}
